/* $Id: f2npoly.c,v 1.7 2007/02/20 14:48:06 dw Exp $ */

/* Arithmetic Library for F2[X] */

/* Copyright (c) Damian Weber <dweber@htw-saarland.de> */
/* License: BSD-style
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "f2npoly.h"
#include "factoring.h"

#include <NTL/ZZ.h>

NTL_CLIENT

#undef DEBUG_MEM
#undef DEBUG

/* given bit position i with 0<=i<=n,
   compute the word and the position within
   the word,
   words are of type ulong 
 */  
void word_bit_pos(int *wp, int *posp, int i)
{
	*wp=i/(sizeof(ulong)*8);
	*posp=i%(sizeof(ulong)*8);
}

/* set the bit at position i */
void set_bit(f2poly_t *fp, int i)
{
	int w; /* which word of f */
	int pos; /* which bit position */
	
	word_bit_pos(&w,&pos,i);
	fp->a[w]|=(unsigned long) 1 <<pos;
	if (i>fp->n)
		fp->n=i;
}

/* negate the bit at position i */
void negate_bit(f2poly_t *fp, int i)
{
	int w; /* which word of f */
	int pos; /* which bit position */
	int j;
	
	word_bit_pos(&w,&pos,i);
	fp->a[w]^=(unsigned long) 1 <<pos;
	if (i>fp->n)
		fp->n=i;

	/* degree has been lowered */
	if (i==fp->n && !is_bit_set(fp,i)) 
	{
		j=fp->n;
		while (j>=0 && is_bit_set(fp,j)==0)
			j--;
		if (j<0) j=0;
		fp->n=j;
	}
}

/* clear the bit at position i */
void clear_bit(f2poly_t *fp, int i)
{
	int w; /* which word of f */
	int pos; /* which bit position */
	
	word_bit_pos(&w,&pos,i);
	fp->a[w]&=~((unsigned long)1<<pos);
}

/* check whether bit at pos i is set */
int is_bit_set(const f2poly_t *fp, int i)
{
	int w; /* which word of f */
	int pos; /* which bit position */
	
	word_bit_pos(&w,&pos,i);
	if (((fp->a[w])&((unsigned long)1<<pos))!=0)
		return 1;
	else
		return 0;
}

/* find MSB of an unsigned long  */
int most_significant_bit(ulong a)
{
	int j;
	ulong mask;

	j=8*sizeof(ulong)-1;
	mask=(ulong)1<<j;
	while ((a&mask)==0 && j>=0)
	{
		mask>>=1;
		j--;
	}
	return j;
}
		
/* shift left by one */
ulong shift_left(ulong a)
{
	return a<<1;
}

/* shift right by one */
ulong shift_right(ulong a)
{
	return a>>1;
}

/* return the degree of f */
int find_leading_bit(const f2poly_t *fp)
{
	return fp->n;
}

/* convert f to string,
   example:
    X^81+X^41+X^5+X+1
*/ 
void to_string(char *str,const f2poly_t *fp)
{
	int i;
	char monom[20]; /* 20 decimal digits enough for 64-bit degree */

	str[0]=0;
	
	i=fp->n;
	while (i>=0)
	{
		if (is_bit_set(fp,i))
		{
			switch(i)
			{
				case 0:
					sprintf(monom,"1");
					break;
				case 1:
					sprintf(monom,"X");
					break;
				default:
					sprintf(monom,"X^%d",i);
					break;
			}
			if (str[0])
				strcat(str,"+");
			strcat(str,monom);
		}
		i--;
	}
	if (!str[0])
		strcpy(str,"0");
}

/* this function must be called for each new ffpoly_t variable
   in order to initialize the pointer a
  */ 

void init_poly(f2poly_t *fp)
{
	fp->a=0;
	fp->n=0;
}

/* return the allocated memory to the system */
void delete_poly(f2poly_t *fp)
{
	if (fp->a)
	{
#ifdef DEBUG_MEM
		cout << " del poly " << (void *) fp->a << endl;
#endif
		delete fp->a;
	}
	init_poly(fp);
}

/* allocate memory for a degree n polynomial */
int new_poly(f2poly_t *fp,int n)
{
	int words;

	if (n<0)
	{ printf("new_poly(): degree %d\n",n);
		return 0;
	}

	if (fp->a)
	{
#ifdef DEBUG_MEM
		cout << " del poly " << (void *) fp->a << endl;
#endif
		delete []fp->a;
	}

	words=n/(sizeof(ulong))+1;
	fp->a=new ulong[words];
	if (!fp->a)
		return 0;
	fp->n=0;
	memset(fp->a,0,words*sizeof(ulong));
#ifdef DEBUG_MEM
	cout << " new poly " << (void *) fp->a << endl;
#endif
	return 1;
}

/* reallocate memory for a degree n polynomial */
int renew_poly(f2poly_t *fp,int n)
{
	int words;
	unsigned int i;
	unsigned int i_bound_f;
	ulong *new_a;

	words=n/(sizeof(ulong))+1;

	new_a=new ulong[words];
	i_bound_f=fp->n/(sizeof(ulong)*8)+1;
	for (i=0;i<i_bound_f;i++)
		new_a[i]=fp->a[i];
	for (i=i_bound_f;i<(unsigned int) words;i++)
		new_a[i]=0;
#ifdef DEBUG_MEM
	cout << " del poly " << (void *) fp->a << endl;
#endif
	delete []fp->a;

	fp->a=new_a;
	if (!fp->a)
		return 0;
	return 1;
}

/* which maximum degree is buried within the input string */
int max_deg_in_string(const char *str)
{
	const char *p;
	int max_deg,deg;

	p=str;
	max_deg=0;
	while ((p=strchr(p,'^'))!=NULL)
	{
		deg=atoi(p+1); /* which degree here after ^ */
		if (deg>max_deg)
			max_deg=deg;
		p++;
	}
	if (max_deg==0) /* no ^ found, degree must be 0 or 1 */
	{
		if (strchr(str,'X'))
			max_deg=1;
		else
			max_deg=0;
	}
	return max_deg;
}

/* convert a string to a polynomial */
void to_f2poly(f2poly_t *fp, const char *str)
{
	const char *p; /* pointer that runs through the string */
	int l;
	int max_deg=0; /* max. deg of a monomial found in input string */
	int monom_deg;

	max_deg=max_deg_in_string(str);
	l=strlen(str);
	if (!new_poly(fp,max_deg))
	{
		puts("could not allocate space for poly");
		return;
	}
	p=str;

	while ((p=strchr(p,'X'))!=NULL) /* find all X powers */
	{
		p++;
		if (*p!='^')
			negate_bit(fp,1);
		else
		{
			p++;
			monom_deg=atoi(p);
			set_bit(fp,monom_deg);
		}
	}
	/* check for constant 1 */
	if (l>=2 && !strncmp(&str[l-2],"+1",2)) /* f(X)=h(X)+1 */
		set_bit(fp,0);
	if (l==1 && str[0]=='1')  /* f(X)=1 */
		set_bit(fp,0);
	fp->n=max_deg;
}

/* set f=g */
int copy_poly(f2poly_t *fp,const f2poly_t *gp)
{
	unsigned int i;

	if (gp==fp) /* if this is the same pointer there is nothing to do */
		return 1;
	
	if (!new_poly(fp,gp->n))
		return 0;
	for (i=0;i<gp->n/(sizeof(ulong)*8)+1;i++)
		fp->a[i]=gp->a[i];
	fp->n=gp->n;
	return 1;
}

/* compute h(X)=f(X)+g(X) */
int add(f2poly_t *hp,const f2poly_t *fp,const f2poly_t *gp)
{
	const f2poly_t *ffp,*ggp; /* ffp will be the one with max degree */
	f2poly_t hh;
	unsigned int i;
	int j;

	
	if (fp->n>gp->n)
	{
		ffp=fp;
		ggp=gp;
	}
	else
	{
		ffp=gp;
		ggp=fp;
	}

	init_poly(&hh);
	if (!new_poly(&hh,ffp->n))
		return 0;
	
	for (i=0;i<=ggp->n/(sizeof(ulong)*8);i++)
		hh.a[i]=ffp->a[i]^ggp->a[i];
	for (i=ggp->n/(sizeof(ulong)*8)+1;i<=ffp->n/(sizeof(ulong)*8);i++)
		hh.a[i]=ffp->a[i];
	if (ggp->n!=ffp->n)
		hh.n=ffp->n;
	else
	{
		/* search leading 1 */
		j=ffp->n;
		while (j>=0 && is_bit_set(&hh,j)==0)
			j--;
		if (j<0) j=0;
		hh.n=j;
	}
	if (!copy_poly(hp,&hh))
		return 0;
	delete_poly(&hh);
	return 1;
}

/* set f(X)=f(X)*X  */
int mul_by_X(f2poly_t *fp)
{
	int w,pos;
	int i;
	ulong carry;

	if ((fp->n+1)%(sizeof(ulong)*8)==0) /* are we at word boundary ? */
	{
		if (!renew_poly(fp,fp->n+1))
			return 0;
	}
	
	word_bit_pos(&w,&pos,fp->n+1);

	carry=0;
	i=w;
	while (i>=0)
	{
		if (i>0)
			carry=(fp->a[i-1] >> (sizeof(ulong)*8-1));
		else
			carry=0;
		fp->a[i]=shift_left(fp->a[i])|carry;
		i--;

	}
	fp->n++;
	return 1;
}

/* set f(X)=f(X)/X, remainder is omitted, assume X|f(X)  */
int div_by_X(f2poly_t *fp)
{
	int w,pos;
	int i;
	ulong carry;

	word_bit_pos(&w,&pos,fp->n);

	i=0;
	while (i<=w)
	{
		if (i==w)
			carry=0;
		else
			carry=fp->a[i+1]&1;
		fp->a[i]=shift_right(fp->a[i])|(carry<<(sizeof(ulong)*8-1));
		i++;
	}
	fp->n--;
	return 1;
}

/* compute h(X)=f(X)*g(X) */
int mul(f2poly_t *hp,const f2poly_t *fp,const f2poly_t *gp)
{
#ifdef DEBUG
	char buffer[2000];
#endif
	f2poly_t g_mul_X;
	f2poly_t hh;
	int i;

	init_poly(&g_mul_X);
	init_poly(&hh);
	if (!copy_poly(&g_mul_X,gp))
		return 0;
	if (!new_poly(&hh,fp->n+gp->n))
		return 0;

	for (i=0;i<=fp->n;i++)
	{
		if (is_bit_set(fp,i))
		{
#if 0
			printf("f has bit %d set\n",i);
			to_string(buffer,&g_mul_X);
			printf("adding    %s    \n",buffer);
 			to_string(buffer,hp);
			printf("to        %s    \n",buffer);
			fflush(stdout);
#endif
			add(&hh,&hh,&g_mul_X);
#if 0
			to_string(buffer,&hh);
			printf("result    %s    \n",buffer);
#endif
		}
		if (!mul_by_X(&g_mul_X))
			return 0;
	}
	copy_poly(hp,&hh);
#ifdef DEBUG
		to_string(buffer,hp);
		cout << "(mul) h=" << buffer << endl;
#endif
	delete_poly(&g_mul_X);
	delete_poly(&hh);
	return 1;
}

/* (slow) power function h(X)=f(X)^m */
int power(f2poly_t *hp,const f2poly_t *fp, int m)
{
	int j;
	f2poly_t prod;

	init_poly(&prod);
	
	if (!new_poly(&prod,m*fp->n))
		return 0;
	prod.a[0]=0x1;
	for (j=1;j<=m;j++)
		if (!mul(&prod,&prod,fp))
			return 0;
	
	if (!copy_poly(hp,&prod))
		return 0;
	
	delete_poly(&prod);
	return 1;
}

/* (fast) power function h(X)=f(X)^m */
int fast_power(f2poly_t *hp,const f2poly_t *fp, ulong m)
{
#ifdef DEBUG
	char buffer[2000];
#endif

	f2poly_t prod;
	f2poly_t q;

	init_poly(&prod);
	init_poly(&q);

	if (!new_poly(&prod,m*fp->n))
		return 0;
	if (!copy_poly(&q,fp));
	return 0;

	prod.a[0]=0x1; /* prod(X)=1 */
	prod.n=0;

	while (m)
	{
		if (m&1)
		{
			if (!mul(&prod,&q,&q))
				return 0;
#ifdef DEBUG
			cout << "(power_mod) MULT" << endl;
#endif
		}
		m>>=1;
		if (m)
		{
			if (!mul(&q,&q,&q))
				return 0;
#ifdef DEBUG
			cout << "(power_mod) SQUARE" << endl;
			to_string(buffer,&q);
			cout << "(power_mod) q=" << buffer << endl;
#endif
		}
#ifdef DEBUG
		to_string(buffer,&prod);
		cout << "(power_mod) prod=" << buffer << endl;
#endif
	}
	
	if (!copy_poly(hp,&prod))
		return 0;
	
	delete_poly(&prod);
	delete_poly(&q);
	return 1;
}

/* (fast) power function h(X)=f(X)^m mod r(X)*/
int fast_power_mod(f2poly_t *hp,const f2poly_t *fp, const ZZ &m0,const f2poly_t *rp)
{
#ifdef DEBUG
	char buffer[2000];
#endif
	f2poly_t prod;
	f2poly_t q;
	f2poly_t quot;

	ZZ m;

	init_poly(&prod);
	init_poly(&q);
	init_poly(&quot);

	if (!new_poly(&prod,1))
		return 0;
	if (!copy_poly(&q,fp))
		return 0;
		
	prod.a[0]=0x1; /* prod(X)=1 */
	prod.n=0;
	m=m0;
	while (m>0)
	{
		if (m % 2 == 1)
		{
			if (!mul(&prod,&prod,&q))
				return 0;
#ifdef DEBUG
			cout << "(power_mod) MULT" << endl;
#endif
		}
		m>>=1;
		if (m>0)
		{
			if (!square_mod(&q,&q,rp))
				return 0;
#ifdef DEBUG
			cout << "(power_mod) SQUARE" << endl;
			to_string(buffer,&q);
			cout << "(power_mod) q=" << buffer << endl;
#endif
		}
		if (!div_rem(&quot,&prod,&prod,rp))
			return 0;
#ifdef DEBUG
		to_string(buffer,&prod);
		cout << "(power_mod) prod=" << buffer << endl;
#endif
	}

	
	if (!copy_poly(hp,&prod))
		return 0;
	
	delete_poly(&prod);
	delete_poly(&q);
	delete_poly(&quot);
	return 1;
}

/* h(X)=f(X)*f(X) mod g(X) */
int square_mod(f2poly_t *hp,const f2poly_t *fp, const f2poly_t *gp)
{
	f2poly_t prod;
	f2poly_t q,r;

	init_poly(&prod);
	init_poly(&q);
	init_poly(&r);

	if (!mul(&prod,fp,fp))
		return 0;

	if (!div_rem(&q,&r,&prod,gp))
		return 0;

	if (!copy_poly(hp,&r))
		return 0;
	
	delete_poly(&prod);
	delete_poly(&q);
	delete_poly(&r);

	return 1;
}

/* divide f by g with quotient q and remainder r */
/* i.e. f(X)=q(X)*g(X)+r(X)                      */
int div_rem(f2poly_t *qp,f2poly_t *rp,const f2poly_t *fp,const f2poly_t *gp)
{
	/* 	char buffer[1000]; */
	int ndiff;
	int nmult;
	f2poly_t gg;
	f2poly_t qq;
	f2poly_t rr;

	init_poly(&gg);
	init_poly(&qq);
	init_poly(&rr);
	if (!new_poly(&qq,fp->n))
		return 0;
	if (!new_poly(&rr,gp->n))
		return 0;
	if (!copy_poly(&gg,gp))
		return 0;
	if (!copy_poly(&rr,fp))
		return 0;

	ndiff=fp->n - gp->n;
	if (ndiff<0) /* quotient is 0, remainder is f */
	{
		if (!new_poly(qp,1))
			return 0;
		qp->n=0;
		qp->a[0]=0;
		if (!copy_poly(rp,fp))
			return 0;
		return 1;
	}
	nmult=0;
	
	while (ndiff>0)
	{
		if (!mul_by_X(&gg))
			return 0;
		ndiff--;
		nmult++;
	}

	/*	printf("# nmult=%d\n",nmult); */
	while (nmult>=0)
	{
		if(rr.n==gg.n)
		{
			add(&rr,&rr,&gg);
			set_bit(&qq,nmult);
#if 0
			to_string(buffer,&rr);
			printf("# rest=%s, degree=%d\n",buffer,rr.n);
#endif	
		}
		if (nmult>0)
		{
			div_by_X(&gg);
#if 0
			to_string(buffer,&gg);
			printf("# gg=%s\n",buffer);
#endif	
		}
		nmult--;
	}
	
	if (!copy_poly(rp,&rr))
		return 0;
	if (!copy_poly(qp,&qq))
		return 0;
	delete_poly(&gg);
	delete_poly(&qq);
	delete_poly(&rr);
	return 1;
}

/* d(X)=gcd(f(X),g(X)) via Euclid */
void gcd(f2poly_t *dp,const f2poly_t *fp,const f2poly_t *gp)
{
	/*	char buffer[1000];*/
	f2poly_t ff;
	f2poly_t gg;
	f2poly_t qq;
	f2poly_t rr;

	init_poly(&ff);
	init_poly(&gg);
	init_poly(&qq);
	init_poly(&rr);
	copy_poly(&ff,fp);
	copy_poly(&gg,gp);

	while (gg.n!=0)
	{
		div_rem(&qq,&rr,&ff,&gg);
		copy_poly(&ff,&gg);
		copy_poly(&gg,&rr);
	}
	if (gg.a[0]==0x1) /* gcd is 1 */
		copy_poly(dp,&gg);
	else		/* gcd is latest divisor */
		copy_poly(dp,&ff);
	delete_poly(&ff);
	delete_poly(&gg);
	delete_poly(&qq);
	delete_poly(&rr);
}


/* test f(X) on irreducibility, if yes return 1
   if no return 0 and a factor of f in g
  */
   
int irred_test(const f2poly_t *fp, f2poly_t *gp)
{
	int k;
	f2poly_t hh; /* X^(p^k) mod f(X) */
	f2poly_t h1; /* X^(p^k)-X mod f(X) */
	f2poly_t X; /* polynomial X */
	f2poly_t dd; /* gcd */

	init_poly(&hh);
	init_poly(&h1);
	init_poly(&dd);
	init_poly(&X);
	new_poly(&X,1);
	X.a[0]=0x2; /* bit 1 is set for X^1 */
	X.n=1;      /* set degree */

	if (!copy_poly(&hh,&X))
	{
		puts("irred_test() error copying polynomial");
		return 0;
	}

	for (k=1;k<=fp->n/2;k++)
	{	/* hh=X^(p^k) mod f(X), with p=2 */
		square_mod(&hh,&hh,fp); 
		add(&h1,&hh,&X); /* h1=X^(p^k)+X */
		gcd(&dd,&h1,fp);
		if (dd.n>=1)
		{
			copy_poly(gp,&dd);
			return 0;
		}
	}
	delete_poly(&hh);
	delete_poly(&h1);
	delete_poly(&dd);
	delete_poly(&X);
	return 1;
}

/* check whether g is a generator of field defined in f,
   m must contain 2^n-1, primes[] must contain the prime factors of m */
int is_generator(const f2poly_t *fp,const f2poly_t *gp,const ZZ &m,const ZZ *primes,int nf)
{
	int i;
	f2poly_t hh;

	init_poly(&hh);
			
	for (i=0;i<=nf;i++)
	{
		if (fast_power_mod(&hh,gp,m/primes[i],fp)==0) /* g^(m/p) */
			return 0;				
		if (hh.n==0 && hh.a[0]==0x1) /* result == 1, no generator */
			return 0;
	}

	delete_poly(&hh);
	return 1;
}
		
/* return 0 on error (field to big, can only ulong size at the moment) */
/* return generator in g (of field defined by f) */
int generator_search(const f2poly_t *fp, f2poly_t *gp)
{
	ZZ m; /* order of the multiplicative group of the finite field */
	ZZ m_tmp; /* m, used for trialdiv */
	int nf;
	ZZ primes[MAX_PRIMEFACTORS];
	int expo[MAX_PRIMEFACTORS];

	if (fp->n<=0)
		return 0;
	if (fp->n>(int)sizeof(ulong)*8-1)
		return 0;

	m=power(to_ZZ(2),fp->n)-1; /* 2^n-1 */
	m_tmp=m;
	trialdiv_ub(m_tmp, 1000000, primes, expo);
	nf=0;
	while (expo[nf]) 
		nf++;
	
	if (!new_poly(gp,fp->n))
		return 0;
	gp->a[0]=0x1;
	gp->n=0;
	if (fp->n==1)
		return 1;

	while (gp->a[0])
	{
		gp->a[0]++; /* next polynomial */
		gp->n=most_significant_bit(gp->a[0]);
		if (is_generator(fp,gp,m,primes,nf))
			return 1;
	}
	return 1;
}


/* search irreducible polynomials of degree deg
   all=0 -> stop after finding the first, store it in f
   all=1 -> print all irreducible polynomials of degree deg
            with 3 or 5 coefficients, return one of them in f
*/

void irred_search(f2poly_t *fp,int deg, int all)
{
	char buffer[1000];
	f2poly_t f,g;
	int i,j,k;

	init_poly(&f);
	init_poly(&g);

	new_poly(&f,deg);
	set_bit(&f,0); /* construct polynomial X^n+X^i+1,X^n+X^i+X^j+X^k+1 */
	set_bit(&f,deg);
	f.n=deg;

	if (deg==0)
	{
		copy_poly(fp,&f);
		if (!all) 
			return;
		puts("1");
		return;
	}
	if (deg==1)
	{
		copy_poly(fp,&f);
		if (!all) 
			return;
		puts("X+1");
		return;
	}


	for (i=1;i<deg;i++)
	{
		set_bit(&f,i);
		if (irred_test(&f,&g))
		{
			copy_poly(fp,&f);
			if (!all)
				return;
			to_string(buffer,&f);
			printf("%s irreducible\n",buffer);
		}
		for (j=i+1;j<deg;j++)
		{
			set_bit(&f,j);
			for (k=j+1;k<deg;k++)
			{
				set_bit(&f,k);
				if (irred_test(&f,&g))
				{
					copy_poly(fp,&f);
					if (!all)
						return;
					to_string(buffer,&f);
					printf("%s irreducible\n",buffer);
				}
				clear_bit(&f,k);
			}
			clear_bit(&f,j);
		}
		clear_bit(&f,i);
	}
	delete_poly(&f);
	delete_poly(&g);
}

