/* filename: plclient.c */
/* pollard-lambda client for F_{2^n} 
 * original version by Michael Walle  20.02.2008
 * extension for ZZ  by Damian Weber
 **/

#include <sstream>
#include <iostream>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "f2npoly.h"

#define GET "GET\n#"
#define PUT "PUT\n"
#define MAX_POLY_LEN 1024
#define MAX_EXP_LEN 1024
#define RECV_BUF_LEN 10240
#define SEND_BUF_LEN 1024
#define MAX_RANDOM_NUMBER_BETA 2048
#define MAX_RANDOM_NUMBER_GAMMA 2048

#define DEBUG_MEM

using namespace std;

struct client {
	struct sockaddr_in server;
	f2poly_t beta;
	f2poly_t gamma;
	f2poly_t poly;
	f2poly_t field_poly;
	int l;
	ZZ exp_beta;
	ZZ exp_gamma;
	ZZ q;
};

void print_maxrss_usage(const char *msg)
{
	struct rusage u;

	getrusage(RUSAGE_SELF,&u);
	cout << "$ maxrss " << msg << " " << u.ru_maxrss << endl;
}


int most_significant_bit(ulong a);

int put_dp(struct client* c)
{
	char poly_buf[MAX_POLY_LEN];
	char buf[SEND_BUF_LEN];

	int sock;

	cout << "PUT_DP" << endl;
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket() failed: ");
		exit(1);	
	}

	if (connect(sock, (struct sockaddr*) &(c->server), sizeof(c->server)) == -1) {
		perror("connect() failed: ");
		exit(1);
	}

	/* send a put message to server */

	c->exp_gamma%=c->q; /* normalize */
	c->exp_beta%=c->q; /* normalize */
	to_string(poly_buf, &(c->poly));


	ostringstream sout;
	sout.str("");
	sout << PUT << "\n" << poly_buf << "\n" << c->exp_gamma 
		<< "\n" << c->exp_beta << "#\n" << ends;
	strncpy(buf, sout.str().c_str(),SEND_BUF_LEN);
	buf[SEND_BUF_LEN-1]=0;
	sout.str("");


	if (send(sock, buf, strlen(buf), 0) == -1) {
		perror("send() failed: ");
		return 0;
	}

	if (shutdown(sock, SHUT_RDWR) == -1) {
		perror("shutdown() failed: ");
		exit(1);
	}

	if (close(sock) == -1) {
		perror("close() failed: ");
		exit(1);
	}

	return 1;
}

int get_polys(struct client* c)
{
	int sock;
	int alen;

	alen=sizeof(c->server);

	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket() failed: ");
		exit(1);	
	}

	if (connect(sock, (struct sockaddr*) &(c->server), alen) == -1) {
		perror("connect() failed: ");
		exit(1);
	}

	/* send a get request to server */
	if (send(sock, GET, sizeof(GET), 0) == -1) {
		perror("send() failed: ");
		return 0;
	}
	
	ssize_t rbytes = 0;
	ssize_t pos = 0;
	char buf[RECV_BUF_LEN];
	memset(buf, 0, RECV_BUF_LEN);

	/* receive polynoms */
	while ((rbytes = recv(sock, buf + pos, RECV_BUF_LEN-pos-1, 0))) {
		pos += rbytes;	
		if (strchr(buf, '#') != 0) {
			break;
		}	
	}
	buf[pos]=0;
	puts("received:");
	puts(buf);

	if (shutdown(sock, SHUT_RDWR) == -1) {
		perror("shutdown() failed: ");
		exit(1);
	}

	if (close(sock) == -1) {
		perror("close() failed: ");
		exit(1);
	}

	char* ppoly = 0;
	if ((ppoly = strtok(buf, "\n")) == 0) {
		printf("parse error: polynom p\n");	
		return 0;
	}
	printf("ppoly=%s\n",ppoly);

	char* pgamma = 0;
	if ((pgamma = strtok(0, "\n")) == 0) {
		printf("parse error: polynom gamma\n");	
		return 0;
	}
	printf("pgamma=%s\n",pgamma);

	char* pbeta = 0;
	if ((pbeta = strtok(0, "\n")) == 0) {
		printf("parse error: polynom beta\n");	
		return 0;
	}
	printf("pbeta=%s\n",pbeta);

	char* pl = 0;
	if ((pl = strtok(0, "\n")) == 0) {
		printf("parse error: distinguished point criteria l\n");	
		return 0;
	}
	printf("pl=%s\n",pl);
	fflush(stdout);


	to_f2poly(&(c->field_poly), ppoly);
	to_f2poly(&(c->gamma), pgamma);
	to_f2poly(&(c->beta), pbeta);
	c->l = atoi(pl);

	return 1;
}

int separate(const f2poly_t* ppoly) {
	 return (ppoly->a[0] & 0xffff) % 3; 
}

void calc(struct client* c) {

#ifdef DEBUG
	char poly_buf[MAX_POLY_LEN];
#endif
	f2poly dummy;

	int err = 1;
	int i;

	init_poly(&dummy);
	/* calculate with polynoms */
	switch(separate(&(c->poly))) {
		case 0 : {
			//printf("C0\n");
			 //cout << "case 0" << endl;
			err = mul(&(c->poly), &(c->poly), &(c->gamma));
			c->exp_gamma++;
			div_rem(&dummy, &(c->poly), &(c->poly), &(c->field_poly));
		} break;
		case 1 : {
			//printf("C1\n");
			 //cout << "case 1" << endl;
			err = mul(&(c->poly), &(c->poly), &(c->beta));
			c->exp_beta++;
			div_rem(&dummy, &(c->poly), &(c->poly), &(c->field_poly));
		} break;
		case 2 : {
			 //cout << "case 2" << endl;
			err = mul(&(c->poly), &(c->poly), &(c->poly));
			c->exp_beta*=2;
			c->exp_gamma*=2;
			if (c->exp_beta>c->q)
				c->exp_beta-=c->q;
			if (c->exp_gamma>c->q)
				c->exp_gamma-=c->q;
			div_rem(&dummy, &(c->poly), &(c->poly), &(c->field_poly));
		} break;
		default: { err = 0; } break;
	}

	if (err == 0) {
		printf("error within calculation\n");
		exit(2);
	}
//	cout << ":: " << c->exp_gamma << " " << c->exp_beta << endl;
}

int is_distinguished_point(f2poly_t* ppoly, int l) {

	/* tupel (i,j) = tupel (array_element, bitpos_within_element) */
	int i = 0;
	int j = 0;

	word_bit_pos(&i, &j, l);
	
	if ((ppoly->a[i] & ((1 << j+1) - 1)) != 0) {
		/* no distinguished point */
		return 0;
	}

	for (i--; i >= 0; i--) {
		if (ppoly->a[i] != 0) {
			/* no distinguished point */
			return 0;
		}
	}

	/* is distinguished point */
	return 1;
}

int main(int argc, char** argv)
{
	char poly_buf[MAX_POLY_LEN];
	f2poly dummy;
	struct client c;	

	if (argc != 3) {
		cout << "usage: " << argv[0] << " <IP> <Port>" << endl;
		return 1;
	}


	init_poly(&(c.beta));
	init_poly(&(c.gamma));
	init_poly(&(c.poly));
	init_poly(&(c.field_poly));

	c.server.sin_family = AF_INET;
	c.server.sin_port = htons(atoi(argv[2]));
	c.server.sin_addr.s_addr = inet_addr(argv[1]);
	
	get_polys(&c);
	init_poly(&(c.poly));
	init_poly(&dummy);

	c.q = power(to_ZZ(2),c.field_poly.n)-1;

	/* select a random start value */
	srandom(time(0));
	c.exp_beta = to_ZZ(random()+getpid()) % c.q;
	c.exp_gamma= to_ZZ(random()+getpid()) % c.q;
	cout << "start exponents " << c.exp_gamma << " " << c.exp_beta << endl;

	fast_power_mod(&(c.poly), &(c.beta), c.exp_beta, &(c.field_poly));
	fast_power_mod(&dummy, &(c.gamma), c.exp_gamma, &(c.field_poly));
	mul(&(c.poly), &(c.poly), &dummy);
	div_rem(&dummy, &(c.poly), &(c.poly), &(c.field_poly));

	to_string(poly_buf, &(c.poly));
	cout << "start poly for lambda-sequence:" << poly_buf << endl;
	cout << "group order q=" << c.q << endl;

	while(1) {
		if(is_distinguished_point(&(c.poly), c.l)) {
			put_dp(&c);	
		}
		calc(&c);
#if 0
		to_string(poly_buf, &(c.poly));
		cout << "# poly :" << poly_buf << endl;
#endif
	}

	return 0;
}

